/* Copyright (c) 1993-2015, NVIDIA CORPORATION. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of NVIDIA CORPORATION nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// compile with: nvcc cuda_transpose.cu -o output_file

#include <stdio.h>
#include <assert.h>

const int TILE_DIM = 32;
const int BLOCK_ROWS = 8;
const int NUM_REPS = 10000;

__global__ void copy(float *odata, const float *idata)
{
	int x = blockIdx.x * TILE_DIM + threadIdx.x;
	int y = blockIdx.y * TILE_DIM + threadIdx.y;
	int width = gridDim.x * TILE_DIM;

	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		odata[(y+j)*width + x] = idata[(y+j)*width + x];
}

__global__ void transposeNaive(float *odata, const float *idata)
{
	int x = blockIdx.x * TILE_DIM + threadIdx.x;
	int y = blockIdx.y * TILE_DIM + threadIdx.y;
	int width = gridDim.x * TILE_DIM;

	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		odata[x*width + (y+j)] = idata[(y+j)*width + x];
}

__global__ void transposeCoalesced(float *odata, const float *idata)
{
	__shared__ float tile[TILE_DIM][TILE_DIM];
	int x = blockIdx.x * TILE_DIM + threadIdx.x;
	int y = blockIdx.y * TILE_DIM + threadIdx.y;
	int width = gridDim.x * TILE_DIM;

	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		tile[threadIdx.x][threadIdx.y + j] = idata[(y+j)*width + x];

	__syncthreads();

	x = blockIdx.y * TILE_DIM + threadIdx.x;
	y = blockIdx.x * TILE_DIM + threadIdx.y;
	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		odata[(y+j)*width + x] = tile[threadIdx.y + j][threadIdx.x];
}


__global__ void transposeNoBankConflicts(float *odata, const float *idata)
{
	__shared__ float tile[TILE_DIM][TILE_DIM+1];
	int x = blockIdx.x * TILE_DIM + threadIdx.x;
	int y = blockIdx.y * TILE_DIM + threadIdx.y;
	int width = gridDim.x * TILE_DIM;

	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		tile[threadIdx.x][threadIdx.y + j] = idata[(y+j)*width + x];

	__syncthreads();

	x = blockIdx.y * TILE_DIM + threadIdx.x;
	y = blockIdx.x * TILE_DIM + threadIdx.y;
	for (int j = 0; j < TILE_DIM; j+= BLOCK_ROWS)
		odata[(y+j)*width + x] = tile[threadIdx.y + j][threadIdx.x];
}

int main(int argc, char **argv)
{
	const int nx = 2048;
	const int ny = 2048;
	const int mem_size = nx*ny*sizeof(float);

	dim3 dimGrid(nx/TILE_DIM, ny/TILE_DIM, 1);
	dim3 dimBlock(TILE_DIM, BLOCK_ROWS, 1);

	int devId = 0;
	if (argc > 1) devId = atoi(argv[1]);

	cudaDeviceProp prop;
	cudaGetDeviceProperties(&prop, devId);
	printf("\nDevice : %s\n", prop.name);
	printf("Matrix size: %d %d, Block size: %d %d, Tile size: %d %d\n", 
			nx, ny, TILE_DIM, BLOCK_ROWS, TILE_DIM, TILE_DIM);
	printf("dimGrid: %d %d %d. dimBlock: %d %d %d\n",
			dimGrid.x, dimGrid.y, dimGrid.z, dimBlock.x, dimBlock.y, dimBlock.z);

	cudaSetDevice(devId);

	float *h_idata = (float*)malloc(mem_size);
	float *h_cdata = (float*)malloc(mem_size);
	float *h_tdata = (float*)malloc(mem_size);

	float *d_idata, *d_cdata, *d_tdata;
	cudaMalloc(&d_idata, mem_size);
	cudaMalloc(&d_cdata, mem_size);
	cudaMalloc(&d_tdata, mem_size);

	// host
	for (int j = 0; j < ny; j++)
		for (int i = 0; i < nx; i++)
			h_idata[j*nx + i] = j*nx + i;


	// device
	cudaMemcpy(d_idata, h_idata, mem_size, cudaMemcpyHostToDevice);

	// events for timing
	cudaEvent_t startEvent, stopEvent;
	cudaEventCreate(&startEvent);
	cudaEventCreate(&stopEvent);
	float ms;

	// ------------
	// time kernels
	// ------------
	printf("%25s%25s\n", "Routine", "Bandwidth (GB/s)");

	// ----
	// copy 
	// ----
	{
		printf("%25s", "copy");
	
		cudaMemset(d_cdata, 0, mem_size);
		cudaEventRecord(startEvent, 0);

		for (int i = 0; i < NUM_REPS; i++)
			copy<<<dimGrid, dimBlock>>>(d_cdata, d_idata);

		cudaEventRecord(stopEvent, 0);
		cudaEventSynchronize(stopEvent);
		cudaEventElapsedTime(&ms, startEvent, stopEvent);
		cudaMemcpy(h_cdata, d_cdata, mem_size, cudaMemcpyDeviceToHost);

		printf("%20.2f\n", 2 * nx * ny * sizeof(float) * 1e-6 * NUM_REPS / ms);
	}



	// --------------
	// transposeNaive 
	// --------------
	{
		printf("%25s", "naive transpose");

		cudaMemset(d_tdata, 0, mem_size);
		cudaEventRecord(startEvent, 0);

		for (int i = 0; i < NUM_REPS; i++)
			transposeNaive<<<dimGrid, dimBlock>>>(d_tdata, d_idata);

		cudaEventRecord(stopEvent, 0);
		cudaEventSynchronize(stopEvent);
		cudaEventElapsedTime(&ms, startEvent, stopEvent);
		cudaMemcpy(h_tdata, d_tdata, mem_size, cudaMemcpyDeviceToHost);

		printf("%20.2f\n", 2 * nx * ny * sizeof(float) * 1e-6 * NUM_REPS / ms);
	}



	// ------------------
	// transposeCoalesced 
	// ------------------
	{
		printf("%25s", "coalesced transpose");

		cudaMemset(d_tdata, 0, mem_size);
		cudaEventRecord(startEvent, 0);

		for (int i = 0; i < NUM_REPS; i++)
			transposeCoalesced<<<dimGrid, dimBlock>>>(d_tdata, d_idata);

		cudaEventRecord(stopEvent, 0);
		cudaEventSynchronize(stopEvent);
		cudaEventElapsedTime(&ms, startEvent, stopEvent);
		cudaMemcpy(h_tdata, d_tdata, mem_size, cudaMemcpyDeviceToHost);

		printf("%20.2f\n", 2 * nx * ny * sizeof(float) * 1e-6 * NUM_REPS / ms);
	}




	// ------------------------
	// transposeNoBankConflicts
	// ------------------------
	{
		printf("%25s", "conflict-free transpose");

		cudaMemset(d_tdata, 0, mem_size);
		cudaEventRecord(startEvent, 0);

		for (int i = 0; i < NUM_REPS; i++)
			transposeNoBankConflicts<<<dimGrid, dimBlock>>>(d_tdata, d_idata);

		cudaEventRecord(stopEvent, 0);
		cudaEventSynchronize(stopEvent);
		cudaEventElapsedTime(&ms, startEvent, stopEvent);
		cudaMemcpy(h_tdata, d_tdata, mem_size, cudaMemcpyDeviceToHost);

		printf("%20.2f\n", 2 * nx * ny * sizeof(float) * 1e-6 * NUM_REPS / ms);
	}

	// cleanup
	cudaEventDestroy(startEvent);
	cudaEventDestroy(stopEvent);
	cudaFree(d_tdata);
	cudaFree(d_cdata);
	cudaFree(d_idata);
	free(h_idata);
	free(h_tdata);
	free(h_cdata);
}
